"""
Implements:
    PrettyDataclass
"""
from dataclasses import fields, is_dataclass
from typing import Any, Dict, Type, TypeVar

from ._pretty_class import PrettyClass
from ._prettyformatter import pformat

Self = TypeVar("Self", bound="PrettyDataclass")


class PrettyDataclass(PrettyClass):
    """
    Base class for creating pretty dataclasses.

    For the full documentation, see:
        https://simpleart.github.io/prettyformatter/PrettyDataclass

        >>> from dataclasses import dataclass
        >>> from typing import List
        >>> 
        >>> 
        >>> @dataclass
        ... class Data(PrettyDataclass):
        ...     data: List[int]
        ... 
        >>> 
        >>> Data(list(range(1000)))
        Data(data=[0, 1, 2, 3, 4, ..., 997, 998, 999])
        >>> 
        >>> 
        >>> @dataclass
        ... class Person(PrettyDataclass):
        ...     name: str
        ...     birthday: str
        ...     phone_number: str
        ...     address: str
        ... 
        >>> 
        >>> Person("Jane Doe", "2001-01-01", "012-345-6789", "123 Sample St.")
        Person(
            name        = "Jane Doe",
            birthday    = "2001-01-01",
            phone_number    = "012-345-6789",
            address     = "123 Sample St.",
        )
    """

    __slots__ = ()

    def __init_subclass__(cls: Type[Self], **kwargs: Any) -> None:
        """Overrides the `__repr__` with itself."""
        # Save the __repr__ directly onto the subclass so that
        # @dataclass will actually notice it.
        cls.__repr__ = cls.__repr__
        return super().__init_subclass__(**kwargs)

    def __pkwargs__(self: Self) -> Dict[str, Any]:
        """
        Implements pretty formatting for dataclasses based on the
        dataclass fields. If the subclass is not a dataclass, does not
        use the dataclass implementation.
        """
        cls = type(self)
        if not is_dataclass(cls):
            raise NotImplementedError
        return {f.name: getattr(self, f.name) for f in fields(cls)}



"""
Implements:
    PrettyClass
"""
from typing import Any, Dict, Tuple, TypeVar

from ._prettyformatter import pformat, pformat_class

Self = TypeVar("Self", bound="PrettyClass")


class PrettyClass:
    """
    Base class for implementing pretty classes.

    Defines `__format__`, `__str__`, and `__repr__` using `pformat`.

    For the full documentation, see:
        https://simpleart.github.io/prettyformatter/PrettyClass

    Dunders
    ---------

    Implement `__pargs__` and/or `__pkwargs__` if the desired `pformat` is
    `cls_name(*args, **kwargs)`.
        >>> class Dog(PrettyClass):
        ...     
        ...     def __init__(self, name, **kwargs):
        ...         self.name = name
        ...         self.attributes = kwargs
        ...     
        ...     def __pargs__(self):
        ...         return (self.name,)
        ...     
        ...     def __pkwargs__(self):
        ...         return self.attributes
        ... 
        >>> Dog("Fido", age=3)
        Dog("Fido", age=3)

    Implement `__pformat__` for more customized `pformat` behavior.
        >>> class PrettyHelloWorld(PrettyClass):
        ...     
        ...     def __pformat__(self, specifier, depth, indent, shorten, json):
        ...         return f"Hello world! Got {specifier!r}, {depth}, {indent}, {shorten}, {json}."
        ... 
        >>> pprint(PrettyHelloWorld())
        Hello world! Got '', 0, 4, True, False.

    JSON serialization
    --------------------

    If the default `__pformat__` is used and `__pargs__` and/or
    `__pkwargs__` is implemented, then JSON serialization is done by
    converting the `args` into a `list` and the `kwargs` into a `dict`.
    If both are given, they are combined using the format of
    `{"class": cls_name, "args": list(args), "kwargs": kwargs}`.

        >>> pprint(Dog("Fido", age=3), json=True)
        {"class": "Dog", "args": ["Fido"], "kwargs": {"age": 3}}
    """

    __slots__ = ()

    def __format__(self: Self, specifier: str) -> str:
        """
        Implements the format specification for `prettyformatter`.
        """
        return pformat(self, specifier)

    def __pargs__(self: Self) -> Tuple[Any, ...]:
        """
        Defines the positional arguments for "cls_name(*args, **kwargs)".
        Raises NotImplementedError, by default, if there are no args.
        """
        raise NotImplementedError

    def __pformat__(self: Self, specifier: str, depth: int, indent: int, shorten: bool, json: bool) -> str:
        """
        Default implementation uses __pargs__ and __pkwargs__.
        """
        cls = type(self)
        try:
            args = cls.__pargs__(self)
        except NotImplementedError:
            args = None
        else:
            if not isinstance(args, tuple):
                raise TypeError("__pargs__ expected a tuple, got " + repr(args))
        try:
            kwargs = cls.__pkwargs__(self)
        except NotImplementedError:
            kwargs = None
        else:
            if not isinstance(kwargs, dict):
                raise TypeError("__pkwargs__ expected a dict, got " + repr(kwargs))
            for name in kwargs:
                if not isinstance(name, str):
                    raise TypeError("__pkwargs__ expected a dict of 'name': value pairs, got " + repr(name))
        with_indent = dict(specifier=specifier, depth=depth, indent=indent, shorten=shorten, json=json)
        if args is None is kwargs:
            formatter = super(PrettyClass, cls).__format__
            if formatter is not object.__format__:
                return formatter(self, specifier)
            else:
                return super().__repr__()
        elif json:
            if args is None:
                return pformat(kwargs, **with_indent)
            elif kwargs is None:
                return pformat(args, **with_indent)
            else:
                return pformat(
                    {
                        "class": cls.__name__,
                        "args": args,
                        "kwargs": kwargs,
                    },
                    **with_indent,
                )
        elif args is None:
            args = ()
        elif kwargs is None:
            kwargs = {}
        return cls.__name__ + pformat_class(args, kwargs, **with_indent)

    def __pkwargs__(self: Self) -> Dict[str, Any]:
        """
        Defines the keyword arguments for "cls_name(*args, **kwargs)".
        Raises NotImplementedError, by default, if there are no kwargs.
        """
        raise NotImplementedError

    def __str__(self: Self) -> str:
        """
        Implements the format specification for `prettyformatter` with
        default parameters.
        """
        return pformat(self)

    def __repr__(self: Self) -> str:
        """
        Implements the format specification for `prettyformatter` with
        default parameters.
        """
        return pformat(self)
